/**
 * @license hzzlxk-inc.
 * @deprecated Rich-Text Component for cross-platforms.
 * @todo RoadMap for next version:
 * 1. optimize for obsolete data initilization of img elements
 * 2. optimize Regex rules in stripOutImgStyles by only one search
 *
 * @author imrock.
 * @version 1.0.1
 */

if (String.prototype.startsWith === undefined) {
  String.prototype.startsWith = function (str) {
    return this.indexOf(str) === 0;
  };
}

const Peacock = (function () {
  /******************************************* 私有成员 *******************************************/
  /** main dom element to display or edit rich-text's html content. */
  const RtBox = document.getElementById('richText');

  const PlatformNames = ['android', 'ios', 'web'];

  const RangeBackupInfo = {
    "sc": RtBox,
    "so": 0,
    "ec": RtBox,
    "eo": 0
  };

  const DEF_PF_NAME = PlatformNames[2];
  const DEF_PF_CMD = 'console.log';

  const IMG_MAX_QUERY_SIDE_PX = 4096 * 4;
  const IMG_MAX_QUERY_AREA_PX = 4096 * 4096;

  const IMG_QUERY_PREFIX = '?x-oss-process=image';

  /** Default Img Raw Size-Attribute-Name Prefix. */
  const DEF_IMG_RAW_SAN_PREFIX = 'image-'

  var currentPlatform = DEF_PF_NAME;

  var imgOnClickCbPfCmd = DEF_PF_CMD;
  var imgOnDeleteCbPfCmd = DEF_PF_CMD;
  var textStylesCbPfCmd = DEF_PF_CMD;

  var imgRawSanPrefix = DEF_IMG_RAW_SAN_PREFIX;

  /**
   * only work if it is NOT undefined, null or blank string.
   * @param {String?} prefix
   */
  function setImgRawSizeNamePrefix(prefix) {
    if (prefix && (prefix = prefix.trim()).length > 0) {
      imgRawSanPrefix = prefix;
    }
  }

  /**
   *
   * @return {String}
   */
  function getImgRawWidAttrName() {
    return imgRawSanPrefix + 'w';
  }

  /**
   *
   * @return {String}
   */
  function getImgRawHeiAttrName() {
    return imgRawSanPrefix + 'h';
  }

  /**
   *
   * @param {String} name platform name for running.
   */
  function setRunningPlatform(name) {
    name = String(name).trim().toLowerCase();
    if (PlatformNames.indexOf(name) !== -1) {
      currentPlatform = name;
    } else {
      console.warn('name is not in PlatformNames, will use "web" as default!');
      currentPlatform = DEF_PF_NAME;
    }
  }

  /**
   * when rich text editor has some thing pasted in, used for edit mode.
   * @param {Event} event
   */
  function onPasteEventHandler(event) {
    event.preventDefault();
    var clipboardData = (event.originalEvent || event).clipboardData;
    var textData;
    // 兼容针对于opera ie等浏览器
    if (!clipboardData) {
      textData = window.clipboardData.getData('text') || '';
      if (textData !== '') {
        if (window.getSelection) {
          // 针对于ie11 10 9 safari
          var newNode = document.createElement('span');
          newNode.innerHTML = textData;
          window.getSelection().getRangeAt(0).insertNode(newNode);
        } else {
          // 兼容ie10 9 8 7 6 5
          document.selection.createRange().pasteHTML(textData);
        }
      }
    } else {
      // 兼容chorme或hotfire
      textData = clipboardData.getData('text/plain') || '';
      if (textData !== '') {
        document.execCommand('insertText', false, textData);
      }
    }
  }

  /**
   *
   * @param {Event} _e unused event param.
   */
  function queryToNotifyStates(event) {
    if (!isOrContainsNode(RtBox, event.target.activeElement)) {
      return;
    }
    var states = [];
    if (document.queryCommandState('bold')) {
      states.push('bold');
    }
    if (document.queryCommandState('italic')) {
      states.push('italic');
    }
    if (document.queryCommandState('underline')) {
      states.push('underline');
    }
    if (document.queryCommandState('strikeThrough')) {
      states.push('strikeThrough');
    }
    // eval(`${textStylesCbPfCmd}("${states.join(',')}")`);
    //android 5.0 does not support ES6 Template Literals (Template Strings)
    eval(textStylesCbPfCmd + '("' + states.join(',') + '")');
  }

  /**
   *
   * @param {MutationRecord[]} mutations
   */
  function editorMutationCallback(mutations) {
    for (var i = 0; i < mutations.length; i++) {
      var mutation = mutations[i];
      var removedNodes = mutation.removedNodes;
      for (var j = 0; j < removedNodes.length; j++) {
        var removed = removedNodes[j];
        var name = removed.nodeName, type = removed.nodeType;
        if (type === Node.ELEMENT_NODE && name === 'IMG') {
          // eval(`${imgOnDeleteCbPfCmd}("${removed.id}")`);
          eval(imgOnDeleteCbPfCmd + '("' + removed.id + '")');
        }
      }
    }
  }

  /**
   *
   * @return {Number} platform unit width of rich-text container.
   */
  function getRtContainerWidth() {
    var containerWidth;
    try {
      containerWidth = window.getComputedStyle(RtBox, null)
        .getPropertyValue('width');
    } catch (e) {
      console.error(e);
      containerWidth = window.screen.width;
    }
    containerWidth = parseInt(containerWidth);
    if (isNaN(containerWidth) || containerWidth <= 0) {
      containerWidth = 240;
    }
    return containerWidth;
  }

  /**
   *
   * @param {HTMLImageElement} img IMG element.
   * @return {String} id string.
   */
  function getOrGenerateIdFor(img) {
    if (img.id.length === 0) {
      img.id = 'i159xxxxxxxxxx'.replace(/x/g, function (c) {
        return (Math.random() * 10 | 0).toString();
      });
    }
    return img.id;
  }

  /**
   *
   * @param {HTMLImageElement} img IMG element.
   * @return {String} raw-src('https://a.b.c/d', 'file://a/b', or 'uploading.png')
   * or just empty string.
   */
  function getRawResUrlFrom(img) {
    var rawUrl = img.getAttribute('raw-url');
    var url = rawUrl;
    if (!url || (url = url.trim()).length === 0) { // it is null, empty or blank
      url = img.src.trim(); // so let's get it from src attribute
    }
    var queryIndex = url.lastIndexOf(IMG_QUERY_PREFIX);
    // it is a remote resource and had been optimized
    if (url.startsWith('http') && queryIndex > 11) {
      // the rawUrl comes from the src attribute of the img
      // or because the external caller accidently passed
      // the assembled url into the rawUrl parameter when
      // inserting a new image.
      // so here we should substring to get raw url.
      url = url.substring(0, queryIndex);
    }
    if (rawUrl !== url) {
      // Reset the correct raw-url attribute.
      img.setAttribute('raw-url', url);
    }
    return url;
  }

  /**
   *
   * @param {HTMLImageElement} img IMG element.
   * @return {Number} image px unit width.
   */
  function getRawPxWidthFrom(img) {
    var attrName = getImgRawWidAttrName();
    var rawWidth = img.getAttribute(attrName);
    if (rawWidth === null) {
      // In later versions, will change this block to just throw Error
      // which will break down in runtime!!!
      console.warn('Do you forgot to set prefix param in Peacock init?');
      console.warn('Get image raw width from ' + attrName + ' failed!');
      console.warn('Trying to get form "innerw" for you...');
      // 尝试性的帮你兼容一下，后续版本可能会被删除！
      rawWidth = img.getAttribute('innerw');
    }

    rawWidth = parseInt(rawWidth);
    if (isNaN(rawWidth) || rawWidth <= 0) {
      console.warn('No available width! using 480 to follow work!');
      rawWidth = 480;
    }
    return rawWidth;
  }

  /**
   *
   * @param {HTMLImageElement} img IMG element.
   * @return {Number} image px unit height.
   */
  function getRawPxHeightFrom(img) {
    var attrName = getImgRawHeiAttrName();
    var rawHeight = img.getAttribute(attrName);
    if (rawHeight === null) {
      // In later versions, will change this block to just throw Error
      // which will break down in runtime!!!
      console.warn('Do you forgot to set prefix param in Peacock init?');
      console.warn('Get image raw height from ' + attrName + ' failed!');
      console.warn('Trying to get form "innerh" for you...');
      // 尝试性的帮你兼容一下，后续版本可能会被删除！
      rawHeight = img.getAttribute('innerh');
    }

    rawHeight = parseInt(rawHeight);
    if (isNaN(rawHeight) || rawHeight <= 0) {
      console.warn('No available height! using 480 to follow work!');
      rawHeight = 480;
    }
    return rawHeight;
  }

  /**
   *
   * @param {HTMLImageElement} img IMG element.
   * @return {Object} image raw info object.
   */
  function generateImageInfoFrom(img) {
    return {
      'id': getOrGenerateIdFor(img),
      'url': getRawResUrlFrom(img),
      'src': img.src,
      'width': getRawPxWidthFrom(img),
      'height': getRawPxHeightFrom(img)
    };
  }

  /**
   *
   * @param {Number} rtcWidth
   * @param {Number} imgHeight
   * @param {Number} rawPxWidth
   * @param {Number} rawPxHeight
   * @return {String} optimized query part for img raw-src.
   */
  function genImgUrlQuery(rtcWidth, imgHeight, rawPxWidth, rawPxHeight) {
    var pixelRatioFloat = window.devicePixelRatio * 1.0;
    var rw = rtcWidth * pixelRatioFloat;  // resized width  原图缩放后的宽
    var rh = imgHeight * pixelRatioFloat; // resized height 原图缩放后的高

    if (rw > rawPxWidth || rh > rawPxHeight) { // 有任何一边大于原图
      rw = rawPxWidth * 1.0; // rw设回原图的宽，不放大
      rh = rawPxHeight * 1.0; // rh设回原图的高，不放大
    }

    if (rh > IMG_MAX_QUERY_SIDE_PX) { // rh如果还大于OSS最大值
      rh = IMG_MAX_QUERY_SIDE_PX;
      rw = rh * rawPxWidth / rawPxHeight; // 再同比缩小rw
    }

    if (rw > IMG_MAX_QUERY_SIDE_PX) { // rw如果还大于OSS最大值
      rw = IMG_MAX_QUERY_SIDE_PX;
      rh = rw * rawPxHeight / rawPxWidth; // 再同比缩小rh
    }

    var area = rw * rh;
    var areaRate = area / IMG_MAX_QUERY_AREA_PX;
    if (areaRate > 1.0) {              // 如果乘积还大于OSS最大值
      rw /= Math.sqrt(areaRate);       // 这里我们只需要对rw求平方根即可
    }

    rw = parseInt(rw); // 取整

    // var query = `${IMG_QUERY_PREFIX}/resize,w_${rw}/quality,Q_80`;
    var query = IMG_QUERY_PREFIX + '/resize,w_' + rw + '/quality,Q_80';
    if (currentPlatform !== PlatformNames[1]) { // 对非ios平台都用webp
      query += '/format,webp';
    }
    return query;
  }

  /**
   *
   * @param {HTMLImageElement} img IMG element.
   * @param {Number} rtcWidth rich-text container width.
   * @return {JSON} image info json object.
   */
  function refreshSrcAndGetImageInfo(img, rtcWidth) {
    var obj = generateImageInfoFrom(img); // used for return
    /** `
     * // 图片标签的像素宽度，按要求先乘以【设备密度比】进行放大
     * var elementPxWidth = obj.width * window.devicePixelRatio;
     * if (elementPxWidth > rtcPxWidth) { // 如果放大后大于容器宽度了，取容器宽度
     *   elementPxWidth = rtcPxWidth;
     * }
     * // 再按原图的高宽比求出图片标签的像素高度
     * var elementPxHeight = elementPxWidth * obj.height / obj.width;
     * // 再对图片标签设置平台高度
     * var elementHeight = elementPxHeight / window.devicePixelRatio;
     * `
     * 下面的代码等效于上面注释中的代码，相当于算法上的优化，因为该方法基本需要在循环中调用。
     */
    var imgWidth = obj.width, imgHeight = obj.height;
    if (imgWidth > rtcWidth) {
      imgHeight = rtcWidth * obj.height / obj.width;
    }
    // 这里使用Math.round安全，因为只是展示尺寸，还未涉及到OSS的取图规则
    imgHeight = Math.round(imgHeight);
    if (img.height != imgHeight) img.height = imgHeight;

    var query = '', newSrc = obj.src;
    if (obj.url.startsWith('http')) { // it is a remote resource
      query = genImgUrlQuery(rtcWidth, imgHeight, obj.width, obj.height);
      newSrc = obj.url + query; // update src with query for img
    } else {
      // TODO local resource(downloading, uploading, error), empty string etc...
      newSrc = obj.url;
    }
    if (img.src !== newSrc) {
      img.src = newSrc;
      obj.src = newSrc;
    }

    return obj;
  }

  /**
   *
   * @param {HTMLImageElement} img IMG element.
   */
  function setupImgOnClickPlatformListener(img) {
    var imgId = getOrGenerateIdFor(img);
    img.onclick = function () {
      // eval(`${imgOnClickCbPfCmd}("${imgId}")`);
      eval(imgOnClickCbPfCmd + '("' + imgId + '")');
    }
  }

  /**
   * check the ancestor node contains or is the descendant node.
   * @param {Node} ancestor
   * @param {Node} descendant
   * @return {Boolean} True if ancestor contains or is descendant, false otherwise.
   */
  function isOrContainsNode(ancestor, descendant) {
    var node = descendant;
    while (node) {
      if (node === ancestor) return true;
      node = node.parentNode;
    }
    return false;
  }

  /**
   *
   * @param {Range} range
   */
  function clearSelectionRangeToAddNew(range) {
    var selection = window.getSelection();
    if (selection !== null) {
      selection.removeAllRanges();
      selection.addRange(range);
    }
  }

  /**
   * @returns {Range} selected range if already have user selection,
   * Or select end of the rich-text editor.
   */
  function prepareSelectedOrEditorEndRange() {
    var selection = window.getSelection(), range;
    if (selection && selection.rangeCount > 0) {
      range = selection.getRangeAt(0);
    } else {
      range = document.createRange();
      var lastChild = RtBox.lastChild;
      if (lastChild === null) {
        lastChild = document.createTextNode('');
        RtBox.appendChild(lastChild);
      }

      range.setStartAfter(lastChild);
      range.collapse(true);
    }

    return range;
  }

  /**
   *
   * @param {Range} range
   * @param {(Node|DocumentFragment)} nodeOrFragment
   */
  function checkRangeToReplaceOrAppend(range, nodeOrFragment) {
    if (isOrContainsNode(RtBox, range.commonAncestorContainer)) {
      // delete whatever is on the range
      range.deleteContents();
      // insert passed in node
      range.insertNode(nodeOrFragment);
    } else {
      RtBox.appendChild(nodeOrFragment);
    }
  }

  /**
   *
   * @param {Range} range
   * @param {Node} node
   */
  function setRangeCollapseToEndOfNode(range, node) {
    // Preserve the selection
    range = range.cloneRange();
    range.setStartAfter(node);
    range.collapse(true);

    clearSelectionRangeToAddNew(range);
  }

  /**
   *
   * @param {Node} node
   */
  function insertNodeAtRangeOrEndOfEditor(node) {
    var range = prepareSelectedOrEditorEndRange();
    checkRangeToReplaceOrAppend(range, node);
    setRangeCollapseToEndOfNode(range, node);
  }

  /**
   *
   * @param {String} html
   */
  function insertHtmlAtRangeOrEndOfEditor(html) {
    var temp = document.createElement("div");
    temp.innerHTML = html;

    var frag = document.createDocumentFragment(), node, lastNode;
    while ((node = temp.firstChild)) {
      lastNode = frag.appendChild(node);
    }

    var range = prepareSelectedOrEditorEndRange();
    checkRangeToReplaceOrAppend(range, frag);
    setRangeCollapseToEndOfNode(range, lastNode);
  }

  /**
   *
   * @param {String} href
   */
  function addCustomCssLink(href) {
    var link = document.createElement('link');
    link.rel = 'stylesheet';
    link.type = 'text/css';
    link.href = href;
    link.media = 'all';
    document.head.appendChild(link);
  }

  /**
   *
   * @param {String} html
   */
  function setRtBoxInnerHtml(html) {
    RtBox.innerHTML = html;
  }

  /**
   *
   * @return {JSON[]} image info json object array.
   */
  function setupAllImgToGetInfoArray() {
    var imgs = RtBox.querySelectorAll('img');

    var imageArray = [];
    var rtcWidth = getRtContainerWidth();
    for (var i = 0; i < imgs.length; i++) {
      var img = imgs[i];
      var image = refreshSrcAndGetImageInfo(img, rtcWidth);
      setupImgOnClickPlatformListener(img);
      imageArray.push(image);
    }
    return imageArray;
  }

  /**
   *
   * @param {Number} rawPxWidth
   * @param {Number} rawPxHeight
   * @param {String} rawUrl
   * @param {String?} id
   * @return {HTMLImageElement} new img element.
   */
  function createImgElement(rawPxWidth, rawPxHeight, rawUrl, id) {
    if (!id || (id = id.trim()).length === 0) {
      id = 'i' + Date.now();
    }

    var newImg = document.createElement('img');
    newImg.id = id;
    newImg.setAttribute('raw-url', rawUrl);
    newImg.setAttribute(getImgRawWidAttrName(), rawPxWidth);
    newImg.setAttribute(getImgRawHeiAttrName(), rawPxHeight);

    return newImg;
  }

  /**
   *
   * @param {HTMLImageElement} img
   * @param {String} rawUrl
   * @param {Number} rawPxWidth
   * @param {Number} rawPxHeight
   */
  function resetImgAttribute(img, rawUrl, rawPxWidth, rawPxHeight) {
    img.setAttribute('raw-url', rawUrl);
    img.setAttribute(getImgRawWidAttrName(), rawPxWidth);
    img.setAttribute(getImgRawHeiAttrName(), rawPxHeight);
  }

  /**
   *
   * @param {String} htmlStr
   * @return {String}
   */
  function stripOutImgSrc(htmlStr) {
    return htmlStr.replace(/(<img(?:\s+?[^>]*?)*?)\s+?src=['"][^>]*?['"]/gim, "$1");
  }

  /**
   *
   * @param {String} htmlStr
   * @return {String}
   */
  function stripOutImgSrcQuery(htmlStr) {
    return htmlStr.replace(/(<img(?:\s+?[^>]*?)*?\s+?src=['"][^?>]*?)\?[^>]*?(['"])/gim, "$1$2");
  }

  /**
   *
   * @returns {String}
   */
  function stripOutImgStyles(htmlStr) {
    // TODO 想办法把下面三个正则优化成一个，因为每一次都是全局搜索，需要走三遍，效率上不是很好。
    htmlStr = htmlStr.replace(/(<img(?:\s+?[^>]*?)*?)\s+?height=['"]?\d*['"]?/gim, "$1");
    htmlStr = htmlStr.replace(/(<img(?:\s+?[^>]*?)*?)\s+?width=['"]?\d*['"]?/gim, "$1");
    htmlStr = htmlStr.replace(/(<img(?:\s+?[^>]*?)*?)\s+?style=['"][^>]*?['"]/gim, "$1");
    return htmlStr
  }





  /******************************************* 公开成员 *******************************************/
  /** exposed singleton of Peacock with public APIs. */
  return {
    /**
     * init funtion for View Mode that should be called when window.onload().
     * @param {String} platform pass one of {@link PlatformNames}, default is 'web'.
     * @param {String} imgOnClickCbCmd your platform callback command for img onclick.
     * @param {String?} prefix img raw size attribute name prefix.
     * @see DEF_PF_CMD
     */
    initForViewMode: function (platform, imgOnClickCbCmd, prefix) {
      setRunningPlatform(platform); // set running platform
      imgOnClickCbPfCmd = String(imgOnClickCbCmd).trim();
      addCustomCssLink('rt-c-view.css');
      setImgRawSizeNamePrefix(prefix);
    },

    /**
     *
     * @param {String} css style sheet string.
     */
    addCustomCssStyle: function (css) {
      var styleSheet = document.createElement("style");
      styleSheet.innerText = css;
      document.head.appendChild(styleSheet);
    },

    /**
     *
     * @param {String} html
     */
    setRichTextHtmlStr: function (html) {
      setRtBoxInnerHtml(html);
      setupAllImgToGetInfoArray();
    },

    /**
     * iOS can use this to get init height for webview.
     * @param {String} html
     * @return {Number} the hole document body's scrollHeight.
     */
    setRichTextToGetHeight: function (html) {
      this.setRichTextHtmlStr(html);
      return document.body.scrollHeight;
    },

    /**
     *
     * @param {String} html
     * @return {JSON[]} image info json object array.
     */
    setRichTextToGetImages: function (html) {
      setRtBoxInnerHtml(html);
      return setupAllImgToGetInfoArray();
    },

    /**
     *
     * @param {String} id
     * @returns {JSON?} image info json object.
     * @see getAllImages
     */
    getImage: function (id) {
      var img = RtBox.querySelector('img#' + id);
      if (img !== null) {
        return generateImageInfoFrom(img);
      } else {
        return null;
      }
    },

    /**
     * @return {JSON[]} image info json object array.
     * @see getImage
     */
    getAllImages: function () {
      var imgs = RtBox.querySelectorAll('img');
      var imageInfoArray = [];
      for (var i = 0; i < imgs.length; i++) {
        var img = imgs[i];
        var image = generateImageInfoFrom(img);
        imageInfoArray.push(image);
      }
      return imageInfoArray;
    },

    /**
     *
     * @returns {String} innerText inside rich text container
     */
    getRawTextOnly: function () {
      return RtBox.innerText;
    },

    /**
     * Get current rich-text html string without any strip.
     * NOTE: It is not suitable for transfer between platforms.
     * @returns {String} innerHTML inside rich text container
     * @see getRichText
     */
    getDirtyRichTextStr: function () {
      return RtBox.innerHTML;
    },

    /**
     *
     * @returns {JSON} RichText object.
     */
    getRichText: function () {
      var dirty = this.getDirtyRichTextStr();
      var data = stripOutImgStyles(dirty);
      // used for inner project
      var compatData = stripOutImgSrcQuery(data);
      data = stripOutImgSrc(data);
      return {
        // raw-text string without any html node
        'rawTxt': this.getRawTextOnly(),
        // all image object list of current rich-text
        'images': this.getAllImages(),
        // compat clean rich-text string, styles are
        // striped out, but retain img src without
        // query part.
        'cleanCp': compatData,
        // clean rich-text string, all styles and
        // src attributes are strip out already.
        'clean': data,
        'dirty': dirty
      };
    },

    /**
     *
     * @return {Number} count
     */
    getImageCount: function () {
      return RtBox.querySelectorAll('img').length;
    },

    /**
     *
     * @return {Number} count
     */
    getTextCount: function () {
      return this.getRawTextOnly().length;
    },



    // -------------------------------- APIs for EDITING Only!. -------------------------------
    /**
     *
     * @param {String} platform
     * @param {String} imgOnClick
     * @param {String} imgOnDelete
     * @param {String} txtOnStyles
     * @param {String?} prefix img size attribute name prefix for transfer.
     */
    initForEditMode: function (platform, imgOnClick, imgOnDelete, txtOnStyles, prefix) {
      this.initForViewMode(platform, imgOnClick, prefix);
      imgOnDeleteCbPfCmd = imgOnDelete;
      textStylesCbPfCmd = txtOnStyles;
      addCustomCssLink('rt-c-edit.css');

      this.setContentEditable(true);

      RtBox.onpaste = onPasteEventHandler;
      var mutationObserver = new MutationObserver(editorMutationCallback);
      mutationObserver.observe(RtBox, { childList: true });
      RtBox.focus();
      document.addEventListener('selectionchange', queryToNotifyStates);
    },

    /**
     *
     * @param {String} placeholder
     */
    setPlaceholder: function (placeholder) {
      RtBox.setAttribute('placeholder', placeholder);
    },

    /**
     * Set editor element's content editable ONLY in Edit Mode.
     * this is Not the command to switch between View and Edit Mode.
     *
     * NOTE: setContentEditable(true) is NOT the init command for edit Mode!
     *
     * @param {Boolean} isEditable false to disable edit, default value is true.
     * @return {Boolean} isEditable after set.
     */
    setContentEditable: function (isEditable) {
      RtBox.contentEditable = String(isEditable);
      return RtBox.contentEditable;
    },

    /**
     * Toggle editor's bold on/off for the selection or at the insertion point.
     * @return {Boolean} is bold on/off after toggle.
     */
    toggleBoldEditing: function () {
      document.execCommand('bold', false, null);
      return document.queryCommandState('bold');
    },

    /**
     * Toggle editor's italics on/off for the selection or at the insertion point.
     * @return {Boolean} is italics on/off after toggle.
     */
    toggleItalicEditing: function () {
      document.execCommand('italic', false, null);
      return document.queryCommandState('italic');
    },

    /**
     * Toggle editor's underline on/off for the selection or at the insertion point.
     * @return {Boolean} is underline on/off after toggle.
     */
    toggleUnderlineEditing: function () {
      document.execCommand('underline', false, null);
      return document.queryCommandState('underline');
    },

    /**
     * Toggle editor's strikethrough on/off for the selection or at the insertion point.
     * @return {Boolean} is strikethrough on/off after toggle.
     */
    toggleStrikeThroughEditing: function () {
      document.execCommand('strikeThrough', false, null);
      return document.queryCommandState('strikeThrough');
    },

    /**
     *
     * @param {String} text
     */
    insertText: function (text) {
      var textNode = document.createTextNode(String(text));
      insertNodeAtRangeOrEndOfEditor(textNode);
    },

    /**
     *
     */
    insertHrElement: function () {
      insertHtmlAtRangeOrEndOfEditor('<hr />');
    },

    /**
     *
     * @param {Number} rawPxWidth
     * @param {Number} rawPxHeight
     * @param {String} rawUrl
     * @param {String?} id
     * @return {JSON} image info json object.
     */
    insertImgElement: function (rawPxWidth, rawPxHeight, rawUrl, id) {
      var newImg = createImgElement(rawPxWidth, rawPxHeight, rawUrl, id);
      var image = refreshSrcAndGetImageInfo(newImg, getRtContainerWidth());
      setupImgOnClickPlatformListener(newImg);
      insertNodeAtRangeOrEndOfEditor(newImg);
      return image;
    },

    /**
     *
     * @param {String} id
     * @param {String} rawUrl
     * @param {Number} rawPxWidth
     * @param {Number} rawPxHeight
     */
    updateImgElement: function (id, rawUrl, rawPxWidth, rawPxHeight) {
      var img = RtBox.querySelector('img#' + id);
      if (img !== null) {
        resetImgAttribute(img, rawUrl, rawPxWidth, rawPxHeight);
        refreshSrcAndGetImageInfo(img, getRtContainerWidth());
      }
    },

    /**
     *
     * @param {String} id
     */
    deleteImgElement: function (id) {
      var img = RtBox.querySelector('img#' + id);
      if (img !== null) {
        img.remove();
      }
    },

    /**
     *
     * @param {String} id
     * @return {Boolean} is highlighted after toggle, always false if
     * no IMG element queried by passed in id.
     */
    toggleImgElementHighlight: function (id) {
      var img = RtBox.querySelector('img#' + id);
      if (img === null) return false;

      img.classList.toggle('selected');
      return img.classList.contains('selected');
    },

    /**
     * Get user current selection or just use the end position of the rt-editor
     * as the range to backup save.
     * @see restoreBackupedRange
     */
    backupCurrentlyRange: function () {
      var range = prepareSelectedOrEditorEndRange();
      RangeBackupInfo.sc = range.startContainer;
      RangeBackupInfo.so = range.startOffset;
      RangeBackupInfo.ec = range.endContainer;
      RangeBackupInfo.eo = range.endOffset;
    },

    /**
     * @see backupCurrentlyRange
     */
    restoreBackupedRange: function () {
      var range = document.createRange();
      range.setStart(RangeBackupInfo.sc, RangeBackupInfo.so);
      range.setEnd(RangeBackupInfo.ec, RangeBackupInfo.eo);

      clearSelectionRangeToAddNew(range);
    }
  }
})();
